import uuid
from typing import Any, List, Optional

from pydantic import BaseModel, ConfigDict, Field


from aigoofusion.chat.messages.role import Role
from aigoofusion.chat.messages.tool_call import ToolCall


class Message(BaseModel):
    """Message class for input to the LLM models."""

    model_config = ConfigDict(extra="forbid")

    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    role: Role
    content: Optional[str] = None
    name: Optional[str] = None
    tool_calls: Optional[List[ToolCall]] = None  # For Message with `assistant` role
    tool_call_id: Optional[str] = None  # For Message with `tool` role
    tool_results: Optional[List[Any]] = None  # For Message with `tool` role
    request_call_id: Optional[str] = None  # For Message with `tool` role

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Ensure tool_results is only used when role is "tool"
        if self.tool_results is not None and self.role != Role.TOOL:
            raise ValueError("tool_results can only be set when role is 'tool'.")

        # Ensure expect tool_results used when role is "tool"
        if not self.tool_results and self.role == Role.TOOL:
            raise ValueError("Expected tool_results when role is 'tool'.")

        # Ensure tool_call_id is only used when role is "tool"
        if self.tool_call_id is not None and self.role != Role.TOOL:
            raise ValueError("tool_call_id can only be set when role is 'tool'.")

        # Ensure tool_calls is only used when role is "assistant"
        if self.tool_calls is not None and self.role != Role.ASSISTANT:
            raise ValueError("tool_calls can only be set when role is 'assistant'.")



from typing import List
from pydantic import BaseModel, ConfigDict, Field

from aigoofusion.chat.messages.message import Message
from aigoofusion.chat.responses.ai_response import AIResponse


class ChatResponse(BaseModel):
    """ChatResponse Class"""
    model_config = ConfigDict(extra='forbid')
    result: AIResponse 
    process: List[Message] = Field(default_factory=list)



from .bedrock_config import BedrockConfig
from .bedrock_formatter import BedrockFormatter
from .bedrock_model import BedrockModel
from .bedrock_pricing_list import BEDROCK_PRICING
from .bedrock_usage_tracker import bedrock_usage_tracker, track_bedrock_usage
from .bedrock_usage import BedrockUsage

__all__ = [
    "BedrockConfig",
    "BedrockFormatter",
    "BedrockModel",
    "BEDROCK_PRICING",
    "bedrock_usage_tracker",
    "track_bedrock_usage",
    "BedrockUsage",
]



import base64
from aigoofusion.flow.node.node import END, START


class WorkflowVisualizer:
    @staticmethod
    def create_mermaid_diagram(workflow) -> str:
        """Generate Mermaid diagram markdown from workflow."""
        nodes = workflow.nodes
        edges = workflow.edges
        
        # Start building the diagram
        mermaid = [
            "graph TD",
            "    %% Node Styles",
            "    %% Nodes",
        ]
        
        # Add all nodes
        for name, node in nodes.items():
            # Style nodes based on their type
            if name == START:
                mermaid.append(f"    {name}([{name}])")
                # mermaid.append(f"    class {name} start")
            elif name == END:
                mermaid.append(f"    {name}([{name}])")
                # mermaid.append(f"    class {name} end")
            else:
                mermaid.append(f"    {name}({name})")
        
        mermaid.append("    %% Edges")
        
        # Add all edges
        for edge in edges:
            source = edge.source
            
            if edge.condition is None:
                # Simple edges
                for target in edge.targets:
                    mermaid.append(f"    {source} --> {target}")
            else:
                # Conditional edges
                for target in edge.targets:
                    mermaid.append(f"    {source} -.->|condition| {target}")
        
        mermaid.append("style START fill:#777EFE,stroke:#4F54AA,color:white")
        mermaid.append("style END fill:#777EFE,stroke:#4F54AA,color:white")

        return "\n".join(mermaid)

    @staticmethod
    def generate_diagram_url(mermaid_code: str) -> str:
        """Generate URL for Mermaid diagram image."""
        graphbytes = mermaid_code.encode("utf8")
        base64_bytes = base64.urlsafe_b64encode(graphbytes)
        base64_string = base64_bytes.decode("ascii")
        return "https://mermaid.ink/img/" + base64_string


import base64
import inspect
from typing import Any, Callable, Dict, List, Optional, Union

from aigoofusion.exception.aigoo_exception import AIGooException
from aigoofusion.flow.edge.edge import Edge
from aigoofusion.flow.node.node import END, START, Node, NodeType
from aigoofusion.flow.state.memory_manager import MemoryManager
from aigoofusion.flow.state.workflow_state import WorkflowState
from aigoofusion.flow.visualizer.visualizer import WorkflowVisualizer


class AIGooFlow:
    def __init__(
        self,
        initial_state: Optional[Dict[str, Any]] = None,
        memory: Optional[MemoryManager] = None,
    ):
        self.nodes: Dict[str, Node] = {
            START: Node(name=START, node_type=NodeType.START),
            END: Node(name=END, node_type=NodeType.END),
        }
        self.edges: List[Edge] = []
        self.state = WorkflowState(initial_state or {})
        self.visualizer = WorkflowVisualizer()
        self.memory = memory

    def validate_workflow(self) -> bool:
        """
        Validate the workflow structure.
        Ensures START has outgoing edges and END has incoming edges.
        """
        start_has_edge = False
        end_has_edge = False

        for edge in self.edges:
            if edge.source == START:
                start_has_edge = True
            if END in edge.targets:
                end_has_edge = True

        if not start_has_edge:
            raise ValueError("Workflow must have at least one edge from START")
        if not end_has_edge:
            raise ValueError("Workflow must have at least one edge to END")

        return True

    def add_node(
        self, name: str, func: Callable, node_type: NodeType = NodeType.FUNCTION
    ) -> None:
        """Add a node to the workflow."""
        if name in (START, END):
            raise ValueError(f"Cannot add node with reserved name {name}")

        # Modify the function to include state access
        original_func = func

        async def wrapped_func(*args, **kwargs):
            sig = inspect.signature(original_func)
            if "state" in sig.parameters:
                kwargs["state"] = self.state

            if inspect.iscoroutinefunction(original_func):
                return await original_func(*args, **kwargs)
            return original_func(*args, **kwargs)

        wrapped_func.__signature__ = inspect.signature(original_func)  # type: ignore
        wrapped_func.__annotations__ = original_func.__annotations__

        sig = inspect.signature(original_func)
        inputs = list(sig.parameters.keys())

        outputs = []
        if sig.return_annotation != inspect.Signature.empty:
            if hasattr(sig.return_annotation, "__annotations__"):
                outputs = list(sig.return_annotation.__annotations__.keys())

        node = Node(
            name=name,
            node_type=node_type,
            func=wrapped_func,
            inputs=inputs,
            outputs=outputs,
        )
        self.nodes[name] = node

    def add_edge(self, source: str, target: str) -> None:
        """Add a direct edge between nodes."""
        if source not in self.nodes:
            raise ValueError(f"Source node '{source}' not found")
        if target not in self.nodes:
            raise ValueError(f"Target node '{target}' not found")

        edge = Edge(source=source, targets=[target])
        self.edges.append(edge)

    def add_conditional_edge(
        self, source: str, targets: Union[str, List[str]], condition: Callable
    ) -> None:
        """Add a conditional edge with multiple possible targets."""
        if source not in self.nodes:
            raise ValueError(f"Source node '{source}' not found")

        target_list = [targets] if isinstance(targets, str) else targets

        for target in target_list:
            if target not in self.nodes:
                raise ValueError(f"Target node '{target}' not found")

        original_condition = condition

        def wrapped_condition(*args, **kwargs):
            result = original_condition(self.state)
            if result not in target_list and result != END:
                raise ValueError(
                    f"Condition returned '{result}' which is not in targets: {target_list}"
                )
            return result

        edge = Edge(source=source, targets=target_list, condition=wrapped_condition)
        self.edges.append(edge)

    def _handle_state_update(self, additional_state, thread_id=None):
        updated_state = (
            self.memory.update_memory(thread_id, additional_state)
            if self.memory and thread_id
            else additional_state
        )
        self.state._update(updated_state)

    async def execute(
        self,
        additional_state: Dict[str, Any],
        thread_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Execute the workflow."""
        try:
            # validate thread_id
            if self.memory and not thread_id:
                raise AIGooException(
                    "`thread_id` required because workflow has memory."
                )

            if additional_state:
                self._handle_state_update(additional_state, thread_id)
                # if self.memory and thread_id:
                #     # using memory
                #     self.state._update(
                #         self.memory.update_memory(thread_id, additional_state)
                #     )
                # else:
                #     self.state._update(additional_state)

            # Validate workflow before execution
            self.validate_workflow()

            nodes_to_process = [START]
            # processed_nodes = set()

            while nodes_to_process:
                current_node = nodes_to_process.pop(0)
                if current_node == END:
                    continue

                if current_node != START:
                    node = self.nodes[current_node]

                    try:
                        func_inputs = {
                            input_name: self.state.get(input_name)
                            for input_name in node.inputs
                            if input_name != "state"
                            and input_name in self.state.get_current()
                        }

                        if node.func:
                            result = await node.func(**func_inputs)
                            # print(f"execute@AIGooFlow result: {result}")
                            if isinstance(result, dict):
                                self._handle_state_update(result, thread_id)
                                # self.state._update(result)
                            elif len(node.outputs) == 1:
                                self._handle_state_update(
                                    {node.outputs[0]: result}, thread_id
                                )
                                # self.state._update({node.outputs[0]: result})

                    except Exception as e:
                        raise AIGooException(
                            f"Error executing node {current_node}: {str(e)}"
                        )

                # Get next nodes
                next_nodes = []
                for edge in self.edges:
                    if edge.source == current_node:
                        if edge.condition is None:
                            next_nodes.extend(edge.targets)
                        else:
                            target = edge.condition()
                            if target and target != END:
                                next_nodes.append(target)

                nodes_to_process.extend(next_nodes)

            return self.state.get_current()
        except Exception as e:
            raise AIGooException(e)

    def get_diagram_code(self) -> str:
        """Get code for the workflow diagram."""
        return self.visualizer.create_mermaid_diagram(self)

    def get_diagram_base64(self):
        """Generate Mermaid diagram base64 for the workflow.

        Display with:
        ```py
        display(Image(url=your_diagram_url))
        ```
        """

        def get_base64(graph):
            graphbytes = graph.encode("utf8")
            base64_bytes = base64.urlsafe_b64encode(graphbytes)
            base64_string = base64_bytes.decode("ascii")
            return base64_string

        mermaid_code = self.get_diagram_code()
        return get_base64(
            f"""
			 {mermaid_code}
			 """
        )

    def get_diagram_url(self):
        """Generate Mermaid diagram url for the workflow.

        Display with:
        ```py
        display(Image(url=your_diagram_url))
        ```
        """

        def get_url(graph):
            graphbytes = graph.encode("utf8")
            base64_bytes = base64.urlsafe_b64encode(graphbytes)
            base64_string = base64_bytes.decode("ascii")
            return "https://mermaid.ink/img/" + base64_string

        mermaid_code = self.get_diagram_code()
        return get_url(
            f"""
			 {mermaid_code}
			 """
        )



import inspect
import re
from typing import Callable, Dict, get_type_hints


class FunctionParser:
    """Extracts metadata from a function."""

    @staticmethod
    def get_function_metadata(func: Callable) -> Dict:
        """Extracts metadata from the given function."""
        docstring = inspect.getdoc(func) or ""
        signature = inspect.signature(func)
        parameters = signature.parameters
        type_hints = get_type_hints(func)

        # Extract short description
        desc_match = re.search(
            r"^(.*?)(?:\n\s*(Args|Parameters):)", docstring, re.DOTALL
        )
        description = desc_match.group(1).strip() if desc_match else docstring

        return {
            "name": func.__name__,
            "description": description,
            "parameters": parameters,
            "type_hints": type_hints,
            "docstring": docstring,
        }



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from aigoofusion.chat.models.model_provider import ModelProvider
from aigoofusion.chat.responses.ai_response import AIResponse


class BaseAIModel(ABC):
    """BaseAIModel Abstract"""

    def __init__(self):
        """Model provider."""
        self.provider: ModelProvider

    @abstractmethod
    def generate(
        self,
        messages: List[Dict[str, Any]],
        tools: Optional[List[Dict[str, Any]]] = None,
        **kwargs,
    ) -> AIResponse:
        pass

    @abstractmethod
    def generate_stream(
        self,
        messages: List[Dict[str, Any]],
        tools: Optional[List[Dict[str, Any]]] = None,
        **kwargs,
    ) -> Any:
        pass



from typing import List, Optional
from pydantic import BaseModel, ConfigDict

from aigoofusion.chat.messages.tool_call import ToolCall


class AIResponse(BaseModel):
    """AIResponse Class"""

    model_config = ConfigDict(extra="forbid")
    content: Optional[str] = None
    tool_calls: Optional[List[ToolCall]] = None



import math
import os
import re
from typing import Dict, List

from aigoofusion.chat.models.bedrock.bedrock_pricing_list import BEDROCK_PRICING
from aigoofusion.chat.models.model_pricing import ModelPricing


class BedrockUsage:
    """
    BedrockUsage class.

    Count bedrock token usage.

    Usage in `BedrockModel`, `bedrock_usage_tracker` and `track_bedrock_usage`
    """

    def __init__(self):
        self.total_request: int = 0
        self.output_tokens: int = 0
        self.input_tokens: int = 0
        self.total_tokens: int = 0
        self.total_cost: float = 0.0
        self.raw_usages: List[Dict[str, int]] = []
        self.pricing: Dict[str, Dict[str, ModelPricing]] = self._load_pricing() or {}

    def __repr__(self) -> str:
        rounded_up = math.ceil(self.total_cost * 1000000) / 1000000
        return (
            f"Total Tokens: {self.total_tokens}\n"
            f"\tInput Tokens: {self.input_tokens}\n"
            f"\tOutput Tokens: {self.output_tokens}\n"
            f"Total Requests: {self.total_request}\n"
            f"Total Cost (USD): ${rounded_up:.6f}"
        )

    def _load_pricing(self) -> Dict[str, Dict[str, ModelPricing]]:
        """
        Load bedrock pricing from dictionary.
        """
        # Price per 1K tokens for different models (USD)
        # https://aws.amazon.com/bedrock/pricing/
        # https://aws.amazon.com/bedrock/pricing/
        # https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/Concepts.RegionsAndAvailabilityZones.html
        pricing_data = BEDROCK_PRICING
        return {
            model: {
                region: ModelPricing(
                    token_input=pricing["input"],
                    token_output=pricing["output"],
                )
                for region, pricing in regions.items()
            }
            for model, regions in pricing_data.items()
        }

    def update(self, model: str, usage: Dict[str, int]) -> None:
        """
        Update usage statistics and calculate price.

        Args:
            model: Model name
            usage: Dictionary containing token counts
        """

        def has_cross_region_inference_id(s):
            return bool(re.match(r"^.{2}\.", s))

        ONE_K = 1000
        AWS_REGION = os.getenv("BEDROCK_AWS_REGION") or ""
        MODEL = model[3:] if has_cross_region_inference_id(model) else model

        self.raw_usages.append(usage)
        usage_dict = vars(usage) if hasattr(usage, "__dict__") else usage
        self.total_request += 1
        self.input_tokens += usage_dict.get("inputTokens", 0)
        self.output_tokens += usage_dict.get("outputTokens", 0)
        self.total_tokens = self.input_tokens + self.output_tokens

        # Calculate price
        if MODEL in self.pricing:
            price_info = self.pricing[MODEL][AWS_REGION]
            input_price = (self.input_tokens / ONE_K) * price_info.token_input
            output_price = (self.output_tokens / ONE_K) * price_info.token_output
            self.total_cost = input_price + output_price
        pass

    def reset(self):
        """
        Reset usage statistics and calculate price.
        """
        self.total_request = 0
        self.input_tokens = 0
        self.output_tokens = 0
        self.total_tokens = 0
        self.total_cost = 0.0
        self.raw_usages = []



import json
import os
import openai
from typing import Any, Dict, List, Optional

from aigoofusion.chat.messages.tool_call import ToolCall
from aigoofusion.chat.models.base_ai_model import BaseAIModel
from aigoofusion.chat.models.model_provider import ModelProvider
from aigoofusion.chat.models.openai.openai_config import OpenAIConfig
from aigoofusion.chat.models.openai.openai_usage import OpenAIUsage
from aigoofusion.chat.models.openai.openai_usage_tracker import track_openai_usage
from aigoofusion.chat.responses.ai_response import AIResponse
from aigoofusion.exception.aigoo_exception import AIGooException


class OpenAIModel(BaseAIModel):
    """
    OpenAIModel class.

    Example:
    ```python
    # Configuration
    config = OpenAIConfig(
            temperature=0.7
    )
    llm = OpenAIModel(model="gpt-4o-mini", config=config)
    ...
    ```

    """

    def __init__(self, model: str, config: OpenAIConfig):
        if not os.getenv("OPENAI_API_KEY"):
            raise AIGooException("Please provide `OPENAI_API_KEY` on your environment!")
        self.provider = ModelProvider.OPENAI
        self.model_name = model
        self.config = config
        self.client = openai.OpenAI(api_key=os.getenv("OPENAI_API_KEY"))
        self.usage_callback = OpenAIUsage()

    @track_openai_usage
    def __call_openai(self, params: dict[str, Any]):
        try:
            response = self.client.chat.completions.create(**params)
            return response
        except Exception as e:
            raise e

    def generate(
        self,
        messages: List[Dict[str, Any]],
        tools: Optional[List[Dict[str, Any]]] = None,
        **kwargs,
    ) -> AIResponse:
        """
        Generate messages.

        Args:
                messages (List[Dict[str, Any]]): _description_
                tools (Optional[List[Dict[str, Any]]], optional): _description_. Defaults to None.

        Raises:
                AIGooChatException: _description_

        Returns:
                AIResponse: _description_
        """
        try:
            params = {
                "model": self.model_name,
                "messages": messages,
                "temperature": self.config.temperature,
                "max_tokens": self.config.max_tokens,
                "top_p": self.config.top_p,
                "frequency_penalty": self.config.frequency_penalty,
                "presence_penalty": self.config.presence_penalty,
                "stream": False,
                **kwargs,
            }

            if tools:
                params["tools"] = [
                    {"type": "function", "function": tool} for tool in tools
                ]
                params["tool_choice"] = "auto"
            print("\nREQUEST:")
            print(json.dumps(params))

            response = self.__call_openai(params)
            print("\nRESPONSE:")
            print(response)

            message = response.choices[0].message
            tool_calls = None
            content = None
            # print(f"usage openai: {response.usage.model_dump()}")
            # token_usage = self.usage_callback.update(self.model_name, response.usage)

            if hasattr(message, "tool_calls") and message.tool_calls:
                tool_calls = [
                    ToolCall(
                        request_call_id=response.id,
                        tool_call_id=tool_call.id,
                        name=tool_call.function.name,
                        arguments=json.loads(tool_call.function.arguments),
                    )
                    for tool_call in message.tool_calls
                ]
            else:
                content = message.content

            return AIResponse(
                content=content,
                tool_calls=tool_calls,
            )

        except Exception as e:
            raise AIGooException(e)

    def generate_stream(
        self,
        messages: List[Dict[str, Any]],
        tools: Optional[List[Dict[str, Any]]] = None,
        **kwargs,
    ) -> Any:
        """
        Generate messages with streaming.

        Note:
                When using stream=True, the response does not include total usage information (usage field with prompt_tokens, completion_tokens, and total_tokens).

                Why?

                \t- In streaming mode, tokens are sent incrementally, so the API doesnt return a single final response that includes token usage.
                \t- If you need token usage, you must track tokens manually or make a separate non-streaming request.

        Args:
                messages (List[Dict[str, Any]]): _description_
                tools (Optional[List[Dict[str, Any]]], optional): _description_. Defaults to None.

        Raises:
                AIGooChatException: _description_

        Returns:
                Any: _description_
        """
        try:
            params = {
                "model": self.model_name,
                "messages": messages,
                "temperature": self.config.temperature,
                "max_tokens": self.config.max_tokens,
                "stream": True,
                **kwargs,
            }

            if tools:
                params["tools"] = [
                    {"type": "function", "function": tool} for tool in tools
                ]
                params["tool_choice"] = "auto"

            return self.client.chat.completions.create(**params)

        except Exception as e:
            raise AIGooException(e)



from typing import Any, Dict, List, Optional

from aigoofusion.chat.messages.message import Message
from aigoofusion.chat.messages.role import Role
from aigoofusion.chat.messages.tool_call import ToolCall
from aigoofusion.chat.models.bedrock.bedrock_formatter import BedrockFormatter
from aigoofusion.chat.models.model_provider import ModelProvider
from aigoofusion.chat.models.openai.openai_formatter import OpenAIFormatter
from aigoofusion.chat.models.model_formatter import ModelFormatter
from aigoofusion.exception.aigoo_exception import AIGooException


class MessageTemp:
    """MessageTemp class. History only per request, not saved to memory."""

    # Register formatter
    _formatters: Dict[ModelProvider, ModelFormatter] = {
        ModelProvider.OPENAI: OpenAIFormatter(),
        ModelProvider.BEDROCK: BedrockFormatter(),
    }

    def __init__(self):
        self.messages: List[Message] = []

    def add_system_message(self, content: str) -> None:
        self.messages.insert(0, Message(role=Role.SYSTEM, content=content))

    def add_user_message(self, id: str, content: str) -> None:
        self.messages.append(Message(id=id, role=Role.USER, content=content))

    def add_assistant_message(
        self,
        id: str,
        content: Optional[str] = None,
        tool_calls: Optional[List[ToolCall]] = None,
    ) -> None:
        # Update request call id by parent
        if tool_calls:
            for tool_call in tool_calls:
                tool_call.request_call_id = id

        self.messages.append(
            Message(id=id, role=Role.ASSISTANT, content=content, tool_calls=tool_calls)
        )

    def add_tool_message(
        self,
        id: str,
        tool_call_id: str,
        name: str,
        result: str,
        provider: ModelProvider,
        request_call_id: Optional[str] = None,
    ) -> None:
        formatter = self._formatters.get(provider)
        if not formatter:
            raise AIGooException(f"Unsupported model provider: {provider}")

        formatter.format_tool_message(
            messages=self.messages,
            id=id,
            tool_call_id=tool_call_id,
            name=name,
            request_call_id=request_call_id,
            result=result,
        )

    def get_messages(self, provider: ModelProvider) -> List[Dict[str, Any]]:
        formatter = self._formatters.get(provider)
        if not formatter:
            raise AIGooException(f"Unsupported model provider: {provider}")
        return [formatter.format_message(msg) for msg in self.messages]

    def get_instance_messages(self) -> List[Message]:
        # return [msg for msg in self.messages if msg.role != Role.SYSTEM]
        return self.messages

    def clear(self) -> None:
        system_message = next(
            (msg for msg in self.messages if msg.role == Role.SYSTEM), None
        )
        self.messages.clear()
        if system_message:
            self.messages.append(system_message)



import inspect
import re
from typing import Any, Dict, List, Union
from aigoofusion.chat.messages.message import Message
from aigoofusion.chat.messages.role import Role
from aigoofusion.chat.models.model_formatter import ModelFormatter


class BedrockFormatter(ModelFormatter):
    """BedrockFormatter.

    Args:
        id (str): _description_
        role (Role): _description_
        content (Optional[str], optional): _description_. Defaults to None.
        tool_calls (Optional[List[ToolCall]], optional): _description_. Defaults to None.
        tool_call_id (Optional[str], optional): _description_. Defaults to None.
        name (Optional[str], optional): _description_. Defaults to None.

    Returns:
        dict: _description_
    """

    def format_message(self, message: Message) -> dict:
        """Formats message into Bedrock's message format.

        BasicResponse:
        ```
        {
            "role": "user | assistant",
            "content": [
                {
                    "text": "string"
                }
            ]
        }
        ```

        ToolRequestResponse:
        ```
        {
            "role": "assistant",
            "content": [
                {
                    "toolUse": {
                        "toolUseId": "tooluse_kZJMlvQmRJ6eAyJE5GIl7Q",
                        "name": "top_song",
                        "input": {
                            "sign": "WZPZ"
                        }
                    }
                }
            ]
        }
        ```

        ToolResultRequest:
        ```
        {
            "role": "user",
            "content": [
                {
                    "toolResult": {
                        "toolUseId": "tooluse_kZJMlvQmRJ6eAyJE5GIl7Q",
                        "content": [
                            {
                                "json": {
                                    "song": "Elemental Hotel",
                                    "artist": "8 Storey Hike"
                                }
                            }
                        ]
                    }
                }
            ]
        }
        ```
        """
        # for tool result (tool message) in bedrock role become `user`
        role = message.role.value if message.role.value != "tool" else "user"
        message_dict: Dict[str, Any] = {
            "role": role,
        }

        if message.content and not message.tool_results and not message.tool_calls:
            message_dict["content"] = [{"text": message.content}]

        if message.tool_results:
            message_dict["content"] = message.tool_results

        if message.tool_calls:
            message_dict["content"] = [
                {
                    "toolUse": {
                        "toolUseId": tool_call.tool_call_id,
                        "name": tool_call.name,
                        "input": tool_call.arguments,
                    }
                }
                for tool_call in message.tool_calls
            ]

        if message.name and not message.role.value == "tool":
            message_dict["name"] = message.name

        return message_dict

    def format_tool_function(
        self, func_metadata: Dict, type_mapping: dict[Any, str]
    ) -> Dict:
        """Formats a function into Bedrock's tool format.

        ```
        {
            "tools": [
                {
                    "toolSpec": {
                        "name": "get_current_weather",
                        "description": "",
                        "inputSchema": {
                            "json": {
                                "type": "object",
                                "properties": {
                                    "location": {
                                        "type": "string",
                                        "description": ""
                                    },
                                    "unit": {
                                        "type": "string",
                                        "description": " (default: celsius)"
                                    }
                                },
                                "required": [
                                    "location",
                                    "unit"
                                ]
                            }
                        }
                    }
                }
            ]
        }
        ```
        """
        metadata = func_metadata

        tool_def = {
            "name": metadata["name"],
            "description": metadata["description"] or metadata["name"],
            "inputSchema": {
                "json": {
                    "type": "object",
                    "properties": {},
                    "required": [],
                }
            },
        }

        for param_name, param in metadata["parameters"].items():
            if param_name == "self":  # Skip 'self' for methods
                continue

            python_type = metadata["type_hints"].get(param_name, param.annotation)
            if hasattr(python_type, "__origin__") and python_type.__origin__ is Union:
                # Extract non-None types from Union
                types = [t for t in python_type.__args__ if t is not type(None)]
                python_type = types[0] if len(types) == 1 else str

            param_type = type_mapping.get(python_type, "string")

            # Extract parameter description
            param_description = ""
            param_patterns = [
                f"{param_name} (",  # Google style
                f"{param_name}:",  # Sphinx style
                f":param {param_name}:",  # reST style
            ]

            for pattern in param_patterns:
                if pattern in metadata["docstring"]:
                    start = metadata["docstring"].find(pattern) + len(pattern)
                    end = metadata["docstring"].find("\n", start)
                    param_description = metadata["docstring"][start:end].strip()
                    if "):" in param_description:
                        match = re.search(r"\):\s*(.*)", param_description)
                        if match:
                            param_description = match.group(1)
                    break

            # Extract default value
            param_default = (
                f"(default: {param.default})"
                if param.default != inspect.Parameter.empty
                else ""
            )

            # Add parameter details
            tool_def["inputSchema"]["json"]["properties"][param_name] = {
                "type": param_type,
                "description": param_description
                + (" " if param_description else "")
                + param_default,
            }

            # Add required params
            tool_def["inputSchema"]["json"]["required"].append(param_name)

        return tool_def

    def format_tool_message(
        self,
        messages: List[Message],
        id: str,
        tool_call_id: str,
        name: str,
        result: str,
        request_call_id: str | None = None,
    ) -> List[Message]:
        tool_result = {
            "toolResult": {
                "toolUseId": tool_call_id,
                "content": [{"json": {"result": result}}],
            }
        }

        # Find existing Bedrock tool response message
        bedrock_message = next(
            (
                msg
                for msg in messages
                if msg.role == Role.TOOL
                and msg.tool_results
                and msg.request_call_id == request_call_id
            ),
            None,
        )

        if bedrock_message:
            # update
            if bedrock_message.tool_results:
                bedrock_message.tool_results.append(tool_result)
        else:
            # add new
            messages.append(
                Message(
                    id=id,
                    role=Role.TOOL,
                    tool_results=[tool_result],
                    request_call_id=request_call_id,
                )
            )
        return messages



from .edge import Edge
from .helper import tools_node
from .node import Node, NodeType, START, END
from .state import WorkflowState, MemoryManager
from .visualizer import WorkflowVisualizer
from .aigoo_flow import AIGooFlow

__all__ = [
    "AIGooFlow",
    "Edge",
    "tools_node",
    "Node",
    "NodeType",
    "START",
    "END",
    "WorkflowState",
    "MemoryManager",
    "WorkflowVisualizer",
]



from enum import Enum


class Role(str, Enum):
	"""Role enum for `Message` class."""
	SYSTEM = "system"
	USER = "user"
	ASSISTANT = "assistant"
	TOOL = "tool"

	def __str__(self):
		return self.value

	def __repr__(self):
		return f"'{self.value}'"



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from aigoofusion.chat.messages.message import Message
# from aigoofusion.chat.tools.function_parser import FunctionParser


class ModelFormatter(ABC):
    """ModelFormatter.

    Register all derivated intances from this class to:
        - `MessageTemp` class at `aigoofusion/chat/messages/message_temp.py`
        - `Tool` class at `aigoofusion/chat/tools/tool.py`

    Args:
        ABC (_type_): _description_
    """

    @abstractmethod
    def format_message(self, message: Message) -> dict:
        pass

    @abstractmethod
    def format_tool_function(
        self, func_metadata: Dict, type_mapping: dict[Any, str]
    ) -> dict:
        pass

    @abstractmethod
    def format_tool_message(
        self,
        messages: List[Message],
        id: str,
        tool_call_id: str,
        name: str,
        result: str,
        request_call_id: Optional[str] = None,
    ) -> List[Message]:
        pass



from typing import Optional

from pydantic import BaseModel


class OpenAIConfig(BaseModel):
	"""Configuration for OpenAIModel."""
	temperature: float = 0.7
	max_tokens: Optional[int] = None
	top_p: float = 1.0
	frequency_penalty: float = 0.0
	presence_penalty: float = 0.0



from .aigoo_chat import AIGooChat
from .messages import Message, MessageTemp, Role, ToolCall
from .tools import ToolRegistry, Tool
from .responses import AIResponse, ChatResponse
from .models import (
    BaseAIModel,
    ModelPricing,
    OpenAIConfig,
    OpenAIModel,
    OPENAI_PRICING,
    track_openai_usage,
    openai_usage_tracker,
    OpenAIUsage,
    BedrockFormatter,
    BedrockConfig,
    BedrockModel,
    BEDROCK_PRICING,
    bedrock_usage_tracker,
    track_bedrock_usage,
    BedrockUsage,
)

__all__ = [
    "AIGooChat",
    "MessageTemp",
    "Message",
    "Role",
    "ToolCall",
    "ToolRegistry",
    "Tool",
    "AIResponse",
    "ChatResponse",
    "BaseAIModel",
    "ModelPricing",
    "OpenAIConfig",
    "OpenAIModel",
    "OPENAI_PRICING",
    "track_openai_usage",
    "openai_usage_tracker",
    "OpenAIUsage",
    "BedrockConfig",
    "BedrockFormatter",
    "BedrockModel",
    "BEDROCK_PRICING",
    "bedrock_usage_tracker",
    "track_bedrock_usage",
    "BedrockUsage",
]



from dataclasses import dataclass
from typing import Callable, List, Optional, Union


@dataclass
class Edge:
    source: str
    targets: Union[str, List[str]]
    condition: Optional[Callable] = None
    
    def __post_init__(self):
        if isinstance(self.targets, str):
            self.targets = [self.targets]


from .openai_config import OpenAIConfig
from .openai_model import OpenAIModel
from .openai_pricing_list import OPENAI_PRICING
from .openai_usage_tracker import track_openai_usage, openai_usage_tracker
from .openai_usage import OpenAIUsage


__all__ = [
    "OpenAIConfig",
    "OpenAIModel",
    "OPENAI_PRICING",
    "track_openai_usage",
    "openai_usage_tracker",
    "OpenAIUsage",
]



from typing import List, Optional
from pydantic import BaseModel


class BedrockConfig(BaseModel):
    """Configuration for BedrockAnthropicModel."""

    temperature: float = 0.7
    max_tokens: Optional[int] = None
    top_p: float = 1.0
    top_k: Optional[int] = None
    stopSequences: Optional[List[str]] = None



# Price per 1M tokens for different models (USD)
# https://platform.openai.com/docs/pricing
OPENAI_PRICING = {
	"gpt-4o": {
		"input": 2.50,
		"output": 10.00
	},
	"gpt-4o-mini": {
		"input": 0.15,
		"output": 0.60
	},
	"gpt-3.5-turbo": {
		"input": 0.05,
		"output": 1.50
	}
}


from copy import deepcopy
from datetime import datetime, timedelta
from threading import Lock, Timer
from typing import Any, Dict, List, Optional


class MemoryManager:
    def __init__(
        self,
        extend_list: bool = False,
        cleanup: bool = True,
        cleanup_time: int = 9000,
    ):
        """MemoryManager

        Args:
            extend_list (bool, optional): Always extended list value.
                If there is key with list value then list will extended with new value.
            cleanup (bool, optional): run cleanup memories after `cleanup_time`
            cleanup_time (int, optional): default clean up time 90000 seconds = 1 day + 1 hour = 86400s + 3600s
        """
        self._memories: Dict[str, Dict[str, Any]] = {}
        self._lock = Lock()
        self.always_extend_list = extend_list
        self._timestamps = {}  # Timestamps to track last usage of threads
        self._cleanup_task = None  # Reference to the cleanup timer
        self._using_cleanup = cleanup
        self._cleanup_time = cleanup_time

    def get_memory(self, thread_id: str) -> Optional[Dict[str, Any]]:
        """Get memory for a specific thread."""
        with self._lock:
            return deepcopy(self._memories.get(thread_id))

    def deep_merge(self, existing: Any, new_data: Any) -> Any:
        """Recursively merge dictionaries and append new unique items to lists."""
        if isinstance(existing, dict) and isinstance(new_data, dict):
            for key, value in new_data.items():
                if key in existing:
                    existing[key] = self.deep_merge(existing[key], value)
                else:
                    existing[key] = deepcopy(value)
        elif isinstance(existing, list) and isinstance(new_data, list):
            # Append only new unique dictionaries (avoid duplicates)
            existing_dicts = {
                frozenset(d.items()) for d in existing if isinstance(d, dict)
            }
            for item in new_data:
                if (
                    isinstance(item, dict)
                    and frozenset(item.items()) not in existing_dicts
                ):
                    existing.append(deepcopy(item))
                elif item not in existing:  # For non-dict items
                    existing.append(deepcopy(item))
        else:
            # If neither a dict nor a list, just overwrite the value
            existing = deepcopy(new_data)

        return existing

    def update_memory(
        self,
        thread_id: str,
        state: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Update memory for a specific thread."""
        with self._lock:
            # Update the last usage timestamp
            self._timestamps[thread_id] = datetime.now()

            if self._using_cleanup:
                # Ensure cleanup task is running
                if self._cleanup_task is None:
                    self._start_cleanup_task(interval=self._cleanup_time)

            # Check thread_id
            if thread_id not in self._memories:
                self._memories[thread_id] = deepcopy(state)
            else:
                if self.always_extend_list:
                    self._memories[thread_id] = self.deep_merge(
                        self._memories[thread_id], state
                    )
                else:
                    self._memories[thread_id].update(deepcopy(state))

            # hah = deepcopy(self._memories.get(thread_id))
            # pprint.pp(f"\nMEMORY: {thread_id}= {self._memories}")
            # print(f"\nCHECK: {self._memories}")
            return self._memories[thread_id]

    def delete_memory(self, thread_id: str) -> None:
        """Delete memory for a specific thread."""
        with self._lock:
            if thread_id in self._memories:
                del self._memories[thread_id]

    def list_threads(self) -> List[str]:
        """List all thread IDs with active memories."""
        with self._lock:
            return list(self._memories.keys())

    def _start_cleanup_task(self, interval: int):
        """Start a periodic task to clean up inactive threads.

        Args:
            interval (int, optional): Interval in seconds for running the cleanup task.
                (default 90000 seconds = 1 day + 1 hour = 86400s + 3600s)
        """

        def cleanup():
            self._remove_inactive_threads(interval)
            # Schedule the next cleanup only if there are threads left
            if self._memories:
                self._cleanup_task = Timer(interval, cleanup)
                self._cleanup_task.start()
            else:
                self._cleanup_task = None

        if self._cleanup_task is None:  # Prevent multiple timers
            self._cleanup_task = Timer(interval, cleanup)
            self._cleanup_task.start()

    def _remove_inactive_threads(self, interval: int):
        print("---- CLEAN-UP RUNNING ----")
        with self._lock:  # Ensure thread-safe access
            now = datetime.now()
            inactive_threads = [
                thread_id
                for thread_id, last_used in self._timestamps.items()
                if now - last_used > timedelta(seconds=interval)
            ]
            for thread_id in inactive_threads:
                del self._memories[thread_id]
                del self._timestamps[thread_id]

            # Stop cleanup task if no threads remain
            if not self._memories:
                self.cleanup_task = None
                print("No threads left. Stopping cleanup task.")
                print("---- CLEAN-UP STOPPED ----")



from .base_ai_model import BaseAIModel
from .model_pricing import ModelPricing
from .openai import (
    OpenAIConfig,
    OpenAIModel,
    OPENAI_PRICING,
    track_openai_usage,
    openai_usage_tracker,
    OpenAIUsage,
)
from .bedrock import (
    BedrockConfig,
    BedrockFormatter,
    BedrockModel,
    BEDROCK_PRICING,
    bedrock_usage_tracker,
    track_bedrock_usage,
    BedrockUsage,
)


__all__ = [
    "BaseAIModel",
    "ModelPricing",
    "OpenAIConfig",
    "OpenAIModel",
    "OPENAI_PRICING",
    "track_openai_usage",
    "openai_usage_tracker",
    "OpenAIUsage",
    "BedrockConfig",
    "BedrockFormatter",
    "BedrockModel",
    "BEDROCK_PRICING",
    "bedrock_usage_tracker",
    "track_bedrock_usage",
    "BedrockUsage",
]



# Price per 1K tokens for different models (USD)
# https://aws.amazon.com/bedrock/pricing/
# https://aws.amazon.com/bedrock/pricing/
# https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/Concepts.RegionsAndAvailabilityZones.html
BEDROCK_PRICING = {
    "anthropic.claude-3-5-sonnet-20240620-v1:0": {
        "us-east-1": {
            "region": "US East (N. Virginia)",
            "input": 0.003,
            "output": 0.015,
        },
        "us-west-2": {
            "region": "US West (Oregon)",
            "input": 0.003,
            "output": 0.015,
        },
    },
    "anthropic.claude-3-5-sonnet-20241022-v2:0": {
        "us-east-1": {
            "region": "US East (N. Virginia)",
            "input": 0.003,
            "output": 0.015,
        },
        "us-west-2": {
            "region": "US West (Oregon)",
            "input": 0.003,
            "output": 0.015,
        },
    },
    "anthropic.claude-3-5-haiku-20241022-v1:0": {
        "us-east-1": {
            "region": "US East (N. Virginia)",
            "input": 0.0008,
            "output": 0.004,
        },
        "us-west-2": {
            "region": "US West (Oregon)",
            "input": 0.0008,
            "output": 0.004,
        },
    },
    "anthropic.claude-3-sonnet-20240229-v1:0": {
        "us-east-1": {
            "region": "US East (N. Virginia)",
            "input": 0.003,
            "output": 0.015,
        },
        "us-west-2": {
            "region": "US West (Oregon)",
            "input": 0.003,
            "output": 0.015,
        },
    },
    "anthropic.claude-3-haiku-20240307-v1:0": {
        "us-east-1": {
            "region": "US East (N. Virginia)",
            "input": 0.00025,
            "output": 0.00125,
        },
        "us-west-2": {
            "region": "US West (Oregon)",
            "input": 0.00025,
            "output": 0.00125,
        },
    },
    "amazon.nova-pro-v1:0": {
        "us-east-1": {
            "region": "US East (N. Virginia)",
            "input": 0.0008,
            "output": 0.0032,
        },
        "us-west-2": {
            "region": "US West (Oregon)",
            "input": 0.0008,
            "output": 0.0032,
        },
    },
    "amazon.nova-lite-v1:0": {
        "us-east-1": {
            "region": "US East (N. Virginia)",
            "input": 0.00006,
            "output": 0.00024,
        },
        "us-west-2": {
            "region": "US West (Oregon)",
            "input": 0.00006,
            "output": 0.00024,
        },
    },
}



from typing import Any, Callable, Dict

from aigoofusion.chat.models.model_formatter import ModelFormatter
from aigoofusion.chat.models.bedrock.bedrock_formatter import BedrockFormatter
from aigoofusion.chat.models.model_provider import ModelProvider
from aigoofusion.chat.models.openai.openai_formatter import OpenAIFormatter
from aigoofusion.chat.tools.function_parser import FunctionParser
from aigoofusion.chat.tools.function_type_mapping import FUNCTION_TYPE_MAPPING
from aigoofusion.exception.aigoo_exception import AIGooException


class Tool:
    """
    Decorator class for creating tool definitions.
    """

    # Register formatter
    _formatters: Dict[ModelProvider, ModelFormatter] = {
        ModelProvider.OPENAI: OpenAIFormatter(),
        ModelProvider.BEDROCK: BedrockFormatter(),
    }

    def __init__(self, strict: bool = True):
        self.strict = strict

    def __call__(self, func: Callable) -> Callable:
        func._is_tool = True  # type: ignore # Mark the function as a tool
        func._tool_strict = self.strict  # type: ignore # Store strict setting. to check: getattr(func, '_tool_strict', True)
        return func

    @staticmethod
    def _get_tool_definition(func: Callable, provider: ModelProvider) -> Dict[str, Any]:
        formatter = Tool._formatters.get(provider)
        if not formatter:
            raise AIGooException(f"Unsupported model provider: {provider}")

        metadata = FunctionParser.get_function_metadata(func)
        return formatter.format_tool_function(metadata, FUNCTION_TYPE_MAPPING)



from .chat import (
    AIGooChat,
    AIResponse,
    ChatResponse,
    Message,
    Role,
    Tool,
    ToolRegistry,
    OpenAIConfig,
    OpenAIModel,
    openai_usage_tracker,
    track_openai_usage,
    BedrockConfig,
    BedrockModel,
    bedrock_usage_tracker,
    track_bedrock_usage,
)

from .exception import AIGooException

from .flow import (
    AIGooFlow,
    Edge,
    tools_node,
    Node,
    WorkflowState,
    MemoryManager,
    START,
    END,
)

__all__ = [
    "AIGooChat",
    "Message",
    "Role",
    "Tool",
    "ToolRegistry",
    "AIResponse",
    "ChatResponse",
    "OpenAIConfig",
    "OpenAIModel",
    "track_openai_usage",
    "openai_usage_tracker",
    "AIGooException",
    "AIGooFlow",
    "Edge",
    "tools_node",
    "Node",
    "START",
    "END",
    "WorkflowState",
    "MemoryManager",
    "BedrockConfig",
    "BedrockModel",
    "bedrock_usage_tracker",
    "track_bedrock_usage",
]



from enum import Enum


class ModelProvider(str, Enum):
	"""ModelProvider enum."""
	OPENAI = "openai"
	BEDROCK = "bedrock"

	def __str__(self):
		return self.value

	def __repr__(self):
		return f"'{self.value}'"



from contextlib import contextmanager
from contextvars import ContextVar
import functools

from aigoofusion.chat.models.openai.openai_usage import OpenAIUsage

# Thread-safe storage for token usage per request
OPENAI_USAGE_TRACKER_VAR = ContextVar("OPENAI_USAGE_TRACKER", default=OpenAIUsage())


@contextmanager
def openai_usage_tracker():
    """OpenAI usage tracker.

    Use this to track token usage on openai.

    Example:
    ```python
    with openai_usage_tracker() as usage:
            result = llm.generate(messages)
            ...
            print(usage)
    ```

    Yields:
            OpenAIUsage: OpenAI usage accumulation.
    """
    usage_tracker = OpenAIUsage()
    OPENAI_USAGE_TRACKER_VAR.set(usage_tracker)  # Store usage_tracker it in the context
    try:
        yield usage_tracker  # Expose tracker to the context
    finally:
        print("")


def track_openai_usage(func):
    """Decorator to wrap `__call_openai` calls on `OpenAIModel`."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        response = func(*args, **kwargs)
        model = args[1][
            "model"
        ]  # args is tuple[OpenAIModel, params] and params contain `model`
        if response.usage:
            usage = response.usage
            usage_tracker = OPENAI_USAGE_TRACKER_VAR.get()
            usage_tracker.update(model=model, usage=usage)
        return response

    return wrapper



import json
import inspect
import re

from typing import Any, Dict, List, Union
from aigoofusion.chat.messages.message import Message
from aigoofusion.chat.messages.role import Role
from aigoofusion.chat.models.model_formatter import ModelFormatter


class OpenAIFormatter(ModelFormatter):
    """OpenAIFormatter

    BasicResponse:
    ```
    {
        role: "developer | user | assistant",
        content: "Write a haiku about recursion in programming.",
    }
    ```

    ToolRequestResponse:
    ```
    [{
        "id": "call_12345xyz",
        "type": "function",
        "function": {
            "name": "get_weather",
            "arguments": "{\"location\":\"Paris, France\"}"
        }
    }]
    ```
    """

    def format_message(self, message: Message) -> dict:
        message_dict: Dict[str, Any] = {
            "role": message.role.value,
        }

        if message.content and not message.tool_results and not message.tool_calls:
            message_dict["content"] = message.content

        if message.tool_results:
            # in openai tool results only one then use first item.
            message_dict["content"] = message.tool_results[0]

        if message.content is not None:
            message_dict["content"] = message.content

        if message.tool_calls:
            message_dict["tool_calls"] = [
                {
                    "id": tool_call.tool_call_id,
                    "type": "function",
                    "function": {
                        "name": tool_call.name,
                        "arguments": json.dumps(tool_call.arguments),
                    },
                }
                for tool_call in message.tool_calls
            ]

        if message.tool_call_id:
            message_dict["tool_call_id"] = message.tool_call_id

        if message.name:
            message_dict["name"] = message.name

        return message_dict

    def format_tool_function(
        self, func_metadata: Dict, type_mapping: dict[Any, str]
    ) -> Dict:
        """Formats a function into OpenAI's tool format.

        ```
        [{
            "type": "function",
            "function": {
                "name": "get_weather",
                "description": "Get current temperature for a given location.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "City and country e.g. Bogotá, Colombia"
                        }
                    },
                    "required": [
                        "location"
                    ],
                    "additionalProperties": False
                },
                "strict": True
            }
        }]
        ```

        """
        metadata = func_metadata
        strict = True

        tool_def = {
            "name": metadata["name"],
            "description": metadata["description"],
            "strict": strict,
            "parameters": {
                "type": "object",
                "properties": {},
                "required": [],
                "additionalProperties": False,
            },
        }

        for param_name, param in metadata["parameters"].items():
            if param_name == "self":  # Skip 'self' for methods
                continue

            python_type = metadata["type_hints"].get(param_name, param.annotation)
            if hasattr(python_type, "__origin__") and python_type.__origin__ is Union:
                # Extract non-None types from Union
                types = [t for t in python_type.__args__ if t is not type(None)]
                python_type = types[0] if len(types) == 1 else str

            param_type = type_mapping.get(python_type, "string")

            # Extract parameter description
            param_description = ""
            param_patterns = [
                f"{param_name} (",  # Google style
                f"{param_name}:",  # Sphinx style
                f":param {param_name}:",  # reST style
            ]

            for pattern in param_patterns:
                if pattern in metadata["docstring"]:
                    start = metadata["docstring"].find(pattern) + len(pattern)
                    end = metadata["docstring"].find("\n", start)
                    param_description = metadata["docstring"][start:end].strip()
                    if "):" in param_description:
                        match = re.search(r"\):\s*(.*)", param_description)
                        if match:
                            param_description = match.group(1)
                    break

            # Extract default value
            param_default = (
                f" (default: {param.default})"
                if param.default != inspect.Parameter.empty
                else ""
            )

            # Add parameter details
            tool_def["parameters"]["properties"][param_name] = {
                "type": param_type,
                "description": param_description
                + (" " if param_description else "")
                + param_default,
            }

            # Add required params
            if strict or param.default == inspect.Parameter.empty:
                tool_def["parameters"]["required"].append(param_name)

        return tool_def

    def format_tool_message(
        self,
        messages: List[Message],
        id: str,
        tool_call_id: str,
        name: str,
        result: str,
        request_call_id: str | None = None,
    ) -> List[Message]:
        messages.append(
            Message(
                id=id,
                role=Role.TOOL,
                tool_call_id=tool_call_id,
                name=name,
                request_call_id=request_call_id,
                tool_results=[result],
            )
        )
        return messages



from typing import Any, Callable, Dict, List

from aigoofusion.chat.models.base_ai_model import BaseAIModel
from aigoofusion.chat.tools.tool import Tool
from aigoofusion.exception.aigoo_exception import AIGooException


class ToolRegistry:
    """
    TollRegistry class.

    Use this ToolRegistry to using tools in `AIGooFlow`.

    Example:
    ```python
    # register tools
    registry = ToolRegistry(['some_tool','other_tool'])

    # execute
    registry.execute_tool(name='some_tool', arguments:{"arg1": "value"})

    ```

    """

    def __init__(
        self,
        funcs: List[Callable],
        model: BaseAIModel,
    ):
        self._tools: Dict[str, Callable] = {}
        self._tool_definitions: Dict[str, Dict[str, Any]] = {}

        for func in funcs:
            if not hasattr(func, "_is_tool"):
                raise AIGooException("Function must be decorated with @Tool")

            tool_def = Tool._get_tool_definition(func, model.provider)
            self._tools[func.__name__] = func
            self._tool_definitions[func.__name__] = tool_def
            # print(f"`{func.__name__}` registered ✅")

    def get_tool_definitions(self) -> List[Dict[str, Any]]:
        """Get all tool definitions registered with this framework."""
        return list(self._tool_definitions.values())

    def execute_tool(self, name: str, arguments: Dict[str, Any]) -> Any:
        """Execute a registered tool."""
        if name not in self._tools:
            raise AIGooException(f"Tool not found: {name}")
        return self._tools[name](**arguments)



from typing import Any, Dict
from pydantic import BaseModel


class ToolCall(BaseModel):
    """ToolCall Class."""

    tool_call_id: str
    request_call_id: str
    name: str
    arguments: Dict[str, Any]



from typing import List

from aigoofusion.chat.messages.message import Message
from aigoofusion.chat.messages.role import Role
from aigoofusion.chat.tools.tool_registry import ToolRegistry


def tools_node(messages: List[Message], registry: ToolRegistry):
    new_messages = messages
    last_message = new_messages[-1]
    if last_message.tool_calls:
        for tool_call in last_message.tool_calls:
            result = registry.execute_tool(
                name=tool_call.name, arguments=tool_call.arguments
            )
            new_messages.append(
                Message(
                    role=Role.TOOL,
                    request_call_id=tool_call.request_call_id,
                    tool_call_id=tool_call.tool_call_id,
                    tool_results=[str(result)],
                    name=tool_call.name,
                )
            )
    return new_messages



import os
from typing import Any, Dict, List

import boto3

from aigoofusion.chat.messages.tool_call import ToolCall
from aigoofusion.chat.models.base_ai_model import BaseAIModel
from aigoofusion.chat.models.bedrock.bedrock_config import (
    BedrockConfig,
)
from aigoofusion.chat.models.bedrock.bedrock_usage_tracker import track_bedrock_usage
from aigoofusion.chat.models.model_provider import ModelProvider
from aigoofusion.chat.responses.ai_response import AIResponse
from aigoofusion.exception.aigoo_exception import AIGooException


class BedrockModel(BaseAIModel):
    def __init__(self, model: str, config: BedrockConfig):
        if not os.getenv("AWS_ACCESS_KEY_ID"):
            raise AIGooException(
                "Please provide `AWS_ACCESS_KEY_ID` on your environment!"
            )
        if not os.getenv("AWS_SECRET_ACCESS_KEY"):
            raise AIGooException(
                "Please provide `AWS_SECRET_ACCESS_KEY` on your environment!"
            )
        if not os.getenv("BEDROCK_AWS_REGION"):
            raise AIGooException(
                "Please provide `BEDROCK_AWS_REGION` on your environment!"
            )

        self.provider = ModelProvider.BEDROCK
        self.model_name = model
        self.config = config
        self.client = boto3.client(
            "bedrock-runtime",
            aws_access_key_id=os.getenv("AWS_ACCESS_KEY_ID"),
            aws_secret_access_key=os.getenv("AWS_SECRET_ACCESS_KEY"),
            region_name=os.getenv("BEDROCK_AWS_REGION"),
        )

    @track_bedrock_usage
    def __call_bedrock(self, params: dict[str, Any]):
        try:
            response = self.client.converse(**params)
            return response
        except Exception as e:
            raise e

    def generate(
        self,
        messages: List[Dict[str, Any]],
        tools: List[Dict[str, Any]] | None = None,
        **kwargs,
    ) -> AIResponse:
        try:
            _system = [msg for msg in messages if msg["role"] == "system"][0]["content"]
            _messages = [msg for msg in messages if msg["role"] != "system"]

            inferences = {
                "temperature": self.config.temperature,
                "maxTokens": self.config.max_tokens,
                "stopSequences": self.config.stopSequences,
                "topP": self.config.top_p,
            }
            # Remove None values
            inference_config = {
                key: value for key, value in inferences.items() if value is not None
            }

            additionals = {
                "top_k": self.config.top_k,
                **kwargs,
            }
            # Remove None values
            additional_config = {
                key: value for key, value in additionals.items() if value is not None
            }

            params = {
                "modelId": self.model_name,
                "messages": _messages,
                "system": _system,
                "inferenceConfig": inference_config,
                "additionalModelRequestFields": additional_config,
            }

            if tools:
                """
                ToolConfig
                {
                    "tools": [
                        {
                            "toolSpec": {
                                "name": "top_song",
                                "description": "Get the most popular song played on a radio station.",
                                "inputSchema": {
                                    "json": {
                                        "type": "object",
                                        "properties": {
                                            "sign": {
                                                "type": "string",
                                                "description": "The call sign for the radio station for which you want the most popular song. Example calls signs are WZPZ and WKRP."
                                            }
                                        },
                                        "required": [
                                            "sign"
                                        ]
                                    }
                                }
                            }
                        }
                    ]
                }
                """
                params["toolConfig"] = {"tools": [{"toolSpec": tool} for tool in tools]}

            response = self.__call_bedrock(params)

            output_message = response["output"]["message"]
            stop_reason = response["stopReason"]
            tool_calls = None
            content = None

            if stop_reason == "tool_use":
                tool_requests = response["output"]["message"]["content"]
                tool_callings = []
                for tool_request in tool_requests:
                    if "toolUse" in tool_request:
                        tool = tool_request["toolUse"]
                        tool_callings.append(
                            ToolCall(
                                request_call_id=response["ResponseMetadata"][
                                    "RequestId"
                                ],
                                tool_call_id=tool["toolUseId"],
                                name=tool["name"],
                                arguments=tool["input"],
                            )
                        )
                tool_calls = tool_callings
            else:
                content = output_message["content"][0]["text"]

            return AIResponse(
                content=content,
                tool_calls=tool_calls,
            )
        except Exception as e:
            raise AIGooException(e)

    def generate_stream(
        self,
        messages: List[Dict[str, Any]],
        tools: List[Dict[str, Any]] | None = None,
        **kwargs,
    ) -> Any:
        try:
            _system = [msg for msg in messages if msg["role"] == "system"][0]["content"]
            _messages = [msg for msg in messages if msg["role"] != "system"]

            inferences = {
                "temperature": self.config.temperature,
                "maxTokens": self.config.max_tokens,
                "stopSequences": self.config.stopSequences,
                "topP": self.config.top_p,
            }
            # Remove None values
            inference_config = {
                key: value for key, value in inferences.items() if value is not None
            }

            additionals = {
                "top_k": self.config.top_k,
                **kwargs,
            }
            # Remove None values
            additional_config = {
                key: value for key, value in additionals.items() if value is not None
            }

            params = {
                "modelId": self.model_name,
                "messages": _messages,
                "system": _system,
                "inferenceConfig": inference_config,
                "additionalModelRequestFields": additional_config,
            }

            if tools:
                params["toolConfig"] = {"tools": [{"toolSpec": tool} for tool in tools]}

            return self.client.converse_stream(**params)

        except Exception as e:
            raise AIGooException(e)



from pydantic import BaseModel


class ModelPricing(BaseModel):
    """ModelPricing Class"""
    token_input: float
    token_output: float

    def __repr__(self) -> str:
        return str(self.model_dump())


from dataclasses import dataclass, field
from enum import Enum
from typing import Callable, List, Optional


class NodeType(Enum):
    START = "start"
    END = "end"
    FUNCTION = "function"
    CONDITIONAL = "conditional"

# Special node identifiers
START = "START"
END = "END"

@dataclass
class Node:
    name: str
    node_type: NodeType
    func: Optional[Callable] = None
    inputs: List[str] = field(default_factory=list)
    outputs: List[str] = field(default_factory=list)


import math
from typing import Dict, List

from aigoofusion.chat.models.openai.openai_pricing_list import OPENAI_PRICING
from aigoofusion.chat.models.model_pricing import ModelPricing

class OpenAIUsage:
	"""
	OpenAIUsage class. 
	
	Count openai token usage. 
	
	Usage in `OpenAIModel`, `openai_usage_tracker` and `track_openai_usage`
	"""

	def __init__(self):
		self.total_request: int = 0
		self.output_tokens: int = 0
		self.input_tokens: int = 0
		self.total_tokens: int = 0
		self.total_cost: float = 0.0
		self.raw_usages: List[Dict[str, int]] = []
		self.pricing: Dict[str, ModelPricing] = self._load_pricing() or {}

	def __repr__(self) -> str:
		rounded_up = math.ceil(self.total_cost * 1000000) / 1000000
		return (
			f"Total Tokens: {self.total_tokens}\n"
			f"\tInput Tokens: {self.input_tokens}\n"
			f"\tOutput Tokens: {self.output_tokens}\n"
			f"Total Requests: {self.total_request}\n"
			f"Total Cost (USD): ${rounded_up:.6f}"
		)

	def _load_pricing(self) -> Dict[str, ModelPricing]:
		"""
		Load openai pricing from dictionary.
		"""
		# Price per 1M tokens for different models (USD)
		# https://platform.openai.com/docs/pricing
		pricing_data = OPENAI_PRICING
		return {
			model: ModelPricing(
				token_input=data['input'],
				token_output=data['output']
			)
			for model, data in pricing_data.items()
			}


	def update(self, model: str, usage: Dict[str, int]) -> None:
		"""
		Update usage statistics and calculate price.
		
		Args:
		    model: Model name
		    usage: Dictionary containing token counts
		"""
		ONE_MILLION = 1000000

		self.raw_usages.append(usage)
		usage_dict = vars(usage) if hasattr(usage, '__dict__') else usage
		self.total_request += 1
		self.input_tokens += usage_dict.get("prompt_tokens", 0)
		self.output_tokens += usage_dict.get("completion_tokens", 0)
		self.total_tokens = self.input_tokens + self.output_tokens

		# Calculate price
		if model in self.pricing:
			price_info = self.pricing[model]
			input_price = (self.input_tokens / ONE_MILLION) * price_info.token_input
			output_price = (self.output_tokens / ONE_MILLION) * price_info.token_output
			self.total_cost = input_price + output_price
		pass

	def reset(self):
		"""
		Reset usage statistics and calculate price.
		"""
		self.total_request = 0
		self.input_tokens = 0
		self.output_tokens = 0
		self.total_tokens = 0
		self.total_cost = 0.0
		self.raw_usages = []



from copy import deepcopy
from typing import Any, Dict, List


class WorkflowState:
    def __init__(
        self,
        initial_state: Dict[str, Any],
    ):
        self._state = initial_state or {}
        self._history: List[Dict[str, Any]] = []

    def get(self, key: str, default: Any = None) -> Any:
        """Get a value from the state."""
        # Return a deep copy to prevent direct modifications
        value = self._state.get(key, default)
        return deepcopy(value)

    def get_current(self) -> Dict[str, Any]:
        """Get the current state."""
        return deepcopy(self._state)

    def _update(self, values: Dict[str, Any]) -> None:
        """Internal method to update state. Only used by Workflow class."""
        self._state.update(deepcopy(values))
        self._history.append(deepcopy(self._state))



from setuptools import setup, find_packages

# Read the README file
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="aigoofusion",
    version="0.1.6",
    packages=find_packages(),
    include_package_data=True,
    description="`AIGooFusion` is a framework for developing applications by large language models (LLMs)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="irufano",
    author_email="irufano.official@gmail.com",
    url="https://github.com/irufano/aigoofusion",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",
)



import string
from typing import Any, Callable, Dict, List, Optional
import uuid

from aigoofusion.chat.messages.message import Message
from aigoofusion.chat.messages.message_temp import MessageTemp
from aigoofusion.chat.messages.role import Role
from aigoofusion.chat.models.base_ai_model import BaseAIModel
from aigoofusion.chat.responses.chat_response import ChatResponse
from aigoofusion.chat.tools.tool import Tool
from aigoofusion.exception.aigoo_exception import AIGooException


class AIGooChat:
    def __init__(
        self,
        llm: BaseAIModel,
        system_message: Optional[str] = None,
        input_variables: Optional[List[str]] = None,
    ):
        self.model: BaseAIModel = llm
        self.messages_temp: MessageTemp = MessageTemp()
        self.system_message = system_message
        self.input_variables = input_variables or []
        self._tools: Dict[str, Callable] = {}
        self._tool_definitions: Dict[str, Dict[str, Any]] = {}

        def _has_variable_placeholder(s):
            return bool(
                list(string.Formatter().parse(s))
                and any(
                    field for _, field, _, _ in string.Formatter().parse(s) if field
                )
            )

        def _extract_variable_names(s):
            return list(
                set([field for _, field, _, _ in string.Formatter().parse(s) if field])
            )

        # Validate that if system message has variable then input variable should not be empty
        if (
            _has_variable_placeholder(self.system_message)
            if self.system_message
            else False
        ) and not self.input_variables:
            variable_names = _extract_variable_names(self.system_message)
            raise AIGooException(
                f"System messages have placeholder variables, so the `input variable` should not be empty. "
                f"Missing input variables: {variable_names}. "
            )

        # Validate input variables
        if self.system_message and self.input_variables:
            # Validate that all required input variables are in kwargs
            variable_names = _extract_variable_names(self.system_message)
            missing_vars = [
                var for var in variable_names if var not in self.input_variables
            ]
            if missing_vars:
                raise AIGooException(
                    f"Missing required input variables: {missing_vars}. "
                    f"Expected variables: {variable_names}."
                )

    def register_tool(self, funcs: List[Callable]) -> None:
        """Register a tool with this framework instance."""
        for func in funcs:
            if not hasattr(func, "_is_tool"):
                raise AIGooException("Function must be decorated with @Tool")

            tool_def = Tool._get_tool_definition(func, self.model.provider)
            self._tools[func.__name__] = func
            self._tool_definitions[func.__name__] = tool_def

    def __get_tool_definitions(self) -> List[Dict[str, Any]]:
        """Get all tool definitions registered with this framework."""
        return list(self._tool_definitions.values())

    def __execute_tool(self, name: str, arguments: Dict[str, Any]) -> Any:
        """Execute a registered tool."""
        if name not in self._tools:
            raise AIGooException(f"Tool not found: {name}")
        return self._tools[name](**arguments)

    def __validate_system_message(self, **kwargs) -> str | None:
        # If we have a system message and input variables
        try:
            final_system_message = self.system_message if self.system_message else ""
            if self.system_message and self.input_variables:
                # Validate that all required input variables are in kwargs
                missing_vars = [
                    var for var in self.input_variables if var not in kwargs
                ]
                if missing_vars:
                    raise AIGooException(
                        f"Missing required input variables: {missing_vars}. "
                        f"Expected variables: {self.input_variables}, "
                        f"Received variables: {list(kwargs.keys())}"
                    )

                # Create a dictionary of variables from kwargs that match input_variables
                format_variables = {
                    var: kwargs.get(var)
                    for var in self.input_variables
                    if var in kwargs
                }

                # Update the system message with the variables
                final_system_message = self.system_message.format(**format_variables)

            return final_system_message
        except KeyError as e:
            raise AIGooException(f"Required variable {e} not found in kwargs")
        except Exception as e:
            raise AIGooException(f"Error formatting system message: {str(e)}")

    def generate(self, messages: List[Message], **kwargs) -> ChatResponse:
        """
        Generate a response based on a list of messages.

        Args:
            messages: List of Message objects to process
            **kwargs: Additional generation parameters

        Returns:
            AIResponse containing the generated response
        """
        try:
            self.messages_temp.clear()

            if self.system_message:
                # Validate system message
                final_system_message = self.__validate_system_message(**kwargs)

                # Add system message to history
                self.messages_temp.add_system_message(
                    final_system_message if final_system_message else ""
                )

            # Add new messages to history
            for message in messages:
                if message.role == Role.USER:
                    self.messages_temp.add_user_message(
                        message.id,
                        message.content if message.content else "",
                    )
                elif message.role == Role.ASSISTANT:
                    self.messages_temp.add_assistant_message(
                        id=message.id,
                        content=message.content,
                        tool_calls=message.tool_calls,
                    )
                elif message.role == Role.TOOL:
                    self.messages_temp.add_tool_message(
                        id=message.id,
                        request_call_id=message.request_call_id,
                        tool_call_id=message.tool_call_id
                        if message.tool_call_id
                        else "",
                        name=message.name if message.name else "",
                        result=message.tool_results[0] if message.tool_results else "",
                        provider=self.model.provider,
                    )

            response = self.model.generate(
                self.messages_temp.get_messages(provider=self.model.provider),
                tools=self.__get_tool_definitions(),
            )

            self.messages_temp.add_assistant_message(
                id=str(uuid.uuid4()),
                content=response.content,
                tool_calls=response.tool_calls,
            )

            return ChatResponse(
                result=response,
                process=self.messages_temp.get_instance_messages(),
            )
        except Exception as e:
            raise AIGooException(e)

    def generate_with_tools(self, messages: List[Message], **kwargs) -> ChatResponse:
        """
        Generate a response based on a list of messages.

        Args:
            messages: List of Message objects to process
            **kwargs: Additional generation parameters

        Returns:
            AIResponse containing the generated response
        """
        try:
            self.messages_temp.clear()

            if self.system_message:
                # Validate system message
                final_system_message = self.__validate_system_message(**kwargs)

                # Add system message to history
                self.messages_temp.add_system_message(
                    final_system_message if final_system_message else ""
                )

            # Add new messages to history
            for message in messages:
                if message.role == Role.USER:
                    self.messages_temp.add_user_message(
                        message.id,
                        message.content if message.content else "",
                    )
                elif message.role == Role.ASSISTANT:
                    self.messages_temp.add_assistant_message(
                        id=message.id,
                        content=message.content,
                        tool_calls=message.tool_calls,
                    )
                elif message.role == Role.TOOL:
                    self.messages_temp.add_tool_message(
                        id=message.id,
                        request_call_id=message.request_call_id,
                        tool_call_id=message.tool_call_id
                        if message.tool_call_id
                        else "",
                        name=message.name if message.name else "",
                        result=message.tool_results[0] if message.tool_results else "",
                        provider=self.model.provider,
                    )

            while True:
                response = self.model.generate(
                    self.messages_temp.get_messages(provider=self.model.provider),
                    tools=self.__get_tool_definitions(),
                )

                if response.tool_calls:
                    self.messages_temp.add_assistant_message(
                        id=str(uuid.uuid4()),
                        tool_calls=response.tool_calls,
                    )

                    for tool_call in response.tool_calls:
                        result = self.__execute_tool(
                            tool_call.name, tool_call.arguments
                        )
                        self.messages_temp.add_tool_message(
                            id=str(uuid.uuid4()),
                            request_call_id=tool_call.request_call_id,
                            tool_call_id=tool_call.tool_call_id,
                            name=tool_call.name,
                            result=str(result),
                            provider=self.model.provider,
                        )
                    continue

                self.messages_temp.add_assistant_message(
                    id=str(uuid.uuid4()),
                    content=response.content,
                    tool_calls=response.tool_calls,
                )

                return ChatResponse(
                    result=response,
                    process=self.messages_temp.get_instance_messages(),
                )
        except Exception as e:
            raise AIGooException(e)

    def generate_stream(self, messages: List[Message], **kwargs) -> Any:
        """
        Generate a streaming response based on a list of messages.

        Args:
            messages: List of Message objects to process
            **kwargs: Additional generation parameters

        Returns:
            Streaming response from the model
        """
        try:
            self.messages_temp.clear()

            if self.system_message:
                # Validate system message
                final_system_message = self.__validate_system_message(**kwargs)

                # Add system message to history
                self.messages_temp.add_system_message(
                    final_system_message if final_system_message else ""
                )

            # Add new messages to history
            for message in messages:
                if message.role == Role.USER:
                    self.messages_temp.add_user_message(
                        message.id,
                        message.content if message.content else "",
                    )
                elif message.role == Role.ASSISTANT:
                    self.messages_temp.add_assistant_message(
                        id=message.id,
                        content=message.content,
                        tool_calls=message.tool_calls,
                    )
                elif message.role == Role.TOOL:
                    self.messages_temp.add_tool_message(
                        id=message.id,
                        request_call_id=message.request_call_id,
                        tool_call_id=message.tool_call_id
                        if message.tool_call_id
                        else "",
                        name=message.name if message.name else "",
                        result=message.tool_results[0] if message.tool_results else "",
                        provider=self.model.provider,
                    )

            return self.model.generate_stream(
                self.messages_temp.get_messages(provider=self.model.provider),
                tools=self.__get_tool_definitions(),
            )
        except Exception as e:
            raise AIGooException(e)

    def clear_messages_temp(self) -> None:
        self.messages_temp.clear()



from contextlib import contextmanager
from contextvars import ContextVar
import functools

from aigoofusion.chat.models.bedrock.bedrock_usage import BedrockUsage

# Thread-safe storage for token usage per request
BEDROCK_USAGE_TRACKER_VAR = ContextVar("BEDROCK_USAGE_TRACKER", default=BedrockUsage())


@contextmanager
def bedrock_usage_tracker():
    """Bedrock usage tracker.

    Use this to track token usage on openai.

    Example:
    ```python
    with bedrock_usage_tracker() as usage:
            result = llm.generate(messages)
            ...
            print(usage)
    ```

    Yields:
            OpenAIUsage: OpenAI usage accumulation.
    """
    usage_tracker = BedrockUsage()
    BEDROCK_USAGE_TRACKER_VAR.set(
        usage_tracker
    )  # Store usage_tracker it in the context
    try:
        yield usage_tracker  # Expose tracker to the context
    finally:
        print("")


def track_bedrock_usage(func):
    """Decorator to wrap `__call_bedrock` calls on `BedrockModel`."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        response = func(*args, **kwargs)
        model = args[1][
            "modelId"
        ]  # args is tuple[BedrockModel, params] and params contain `modelId`
        if response["usage"]:
            usage = response["usage"]
            usage_tracker = BEDROCK_USAGE_TRACKER_VAR.get()
            usage_tracker.update(model=model, usage=usage)
        return response

    return wrapper
